"""Test the account response."""

from datetime import time

from hcb_soap_client.account_response import AccountResponse, Student, TimeOfDay
from tests import read_file


def test_account_response_from_dict() -> None:
    """Test the from text."""
    response_text = read_file("s1157.xml")
    account_response = AccountResponse.from_text(response_text)
    assert account_response.account_id == "55503A65D2F448AC8DB5F8181B305281"
    expected_students = 2
    assert len(account_response.students) == expected_students
    assert account_response.students[0] == Student(
        student_id="DC90F9C55C6142519C9CC9EFA1847922",
        first_name="Test1",
        last_name="Student1",
    )
    assert account_response.students[1] == Student(
        student_id="5A2AECC78C2E43FD9CC03CCE55D9806D",
        first_name="Test2",
        last_name="Student2",
    )
    expected_times = 3
    assert len(account_response.times) == expected_times
    assert account_response.times[0] == TimeOfDay(
        id="55632A13-35C5-4169-B872-F5ABDC25DF6A",
        name="LookupItem_RouteTimeOfDayType_AM",
        begin_time=time(hour=0, minute=0, second=0),
        end_time=time(hour=10, minute=0, second=0),
    )
    assert account_response.times[1] == TimeOfDay(
        id="27AADCA0-6D7E-4247-A80F-7847C448EEED",
        name="LookupItem_RouteTimeOfDayType_MID",
        begin_time=time(hour=10, minute=0, second=0),
        end_time=time(hour=13, minute=30, second=0),
    )
    assert account_response.times[2] == TimeOfDay(
        id="6E7A050E-0295-4200-8EDC-3611BB5DE1C1",
        name="LookupItem_RouteTimeOfDayType_PM",
        begin_time=time(hour=13, minute=30, second=0),
        end_time=time(hour=23, minute=59, second=59),
    )



"""Init file for tests."""

from pathlib import Path


def read_file(file_name: str) -> str:
    """Read a text file."""
    with Path(f"tests/test_data/{file_name}").open() as file:
        return file.read()



"""Class for the account info."""

from datetime import time
from typing import Annotated, Self

from lxml import etree
from pydantic import BaseModel, BeforeValidator

from . import parse_time_str, xpath_attr, xpath_elements

# Type aliases for cleaner field definitions
TimeStr = Annotated[time, BeforeValidator(parse_time_str)]


class Student(BaseModel):
    """Info for the student."""

    student_id: str
    first_name: str
    last_name: str

    @classmethod
    def from_element(cls, elem: etree._Element) -> "Student":
        """Create from lxml element."""
        return cls(
            student_id=elem.get("EntityID", ""),
            first_name=elem.get("FirstName", ""),
            last_name=elem.get("LastName", ""),
        )


class TimeOfDay(BaseModel):
    """The time of day list."""

    id: str
    name: str
    begin_time: TimeStr
    end_time: TimeStr

    @classmethod
    def from_element(cls, elem: etree._Element) -> "TimeOfDay":
        """Create from lxml element."""
        return cls(
            id=elem.get("ID", ""),
            name=elem.get("Name", ""),
            begin_time=elem.get("BeginTime", "00:00:00"),
            end_time=elem.get("EndTime", "00:00:00"),
        )


class AccountResponse(BaseModel):
    """Parent account info."""

    account_id: str
    students: list[Student]
    times: list[TimeOfDay]

    @classmethod
    def from_text(cls, response_text: str) -> Self:
        """Create a new instance from text."""
        root = etree.fromstring(response_text.encode())

        account_id = xpath_attr(root, "//*[local-name()='Account']/@ID")
        student_elements = xpath_elements(root, "//*[local-name()='Student']")
        students = [Student.from_element(e) for e in student_elements]
        time_elements = xpath_elements(root, "//*[local-name()='TimeOfDay']")
        times = [TimeOfDay.from_element(e) for e in time_elements]

        return cls(account_id=account_id, students=students, times=times)



"""Run a test against the live api."""

import os

import pytest

from hcb_soap_client.hcb_soap_client import HcbSoapClient


@pytest.mark.skip(reason="Live testing disabled.")
async def test_async() -> None:
    """Run the async test process."""
    client = HcbSoapClient()
    school_code = os.environ["HCB_SCHOOLCODE"]
    user_name = os.environ["HCB_USERNAME"]
    password = os.environ["HCB_PASSWORD"]
    school_id = await client.get_school_id(school_code)
    if school_id == "":
        msg = "school_id was blank."
        raise ValueError(msg)
    parent_info = await client.get_parent_info(school_id, user_name, password)
    if parent_info.account_id == "":
        msg = "account_id was blank."
        raise ValueError(msg)
    if len(parent_info.students) == 0:
        msg = "No students found."
        raise ValueError(msg)
    student_id = parent_info.students[0].student_id
    if student_id == "":
        msg = "student_id was blank."
        raise ValueError(msg)
    stops = await client.get_stop_info(
        school_id, parent_info.account_id, student_id, HcbSoapClient.PM_ID
    )
    if len(stops.student_stops) == 0:
        msg = "No stops found."
        raise ValueError(msg)



"""Init file for the soap client."""

from datetime import time

from lxml import etree

# Public API exports - must be after utility functions are defined
# to avoid circular imports, so we import at module level but define
# utilities first inline

_MIN_PARTS_WITH_SECONDS = 3


def xpath_attr(root: etree._Element, expr: str) -> str:
    """Get first attribute result from XPath expression, or empty string."""
    result = root.xpath(expr)
    return result[0] if result else ""


def xpath_element(root: etree._Element, expr: str) -> etree._Element | None:
    """Get first element from XPath expression, or None."""
    result = root.xpath(expr)
    return result[0] if result else None


def xpath_elements(root: etree._Element, expr: str) -> list[etree._Element]:
    """Get elements from XPath expression."""
    return root.xpath(expr)


def parse_time_str(value: str | time) -> time:
    """Parse time string (HH:MM:SS) to time object."""
    if isinstance(value, time):
        return value
    parts = value.split(":")
    return time(
        hour=int(parts[0]),
        minute=int(parts[1]),
        second=int(parts[2]) if len(parts) >= _MIN_PARTS_WITH_SECONDS else 0,
    )


def parse_yn_bool(value: str) -> bool:
    """Parse Y/N or Yes/No string to bool."""
    return str(value).upper().startswith("Y")


# ruff: noqa: E402
# Public API exports - placed after utility definitions to avoid circular imports
from .account_response import AccountResponse
from .hcb_soap_client import HcbApiError, HcbSoapClient
from .stop_response import StopResponse

__all__ = [
    "AccountResponse",
    "HcbApiError",
    "HcbSoapClient",
    "StopResponse",
]



"""Cost used for test data."""

SCHOOL_ID = "5C9756CEEA1B438A80C65C20D2720E23"
ACCOUNT_ID = "55503A65D2F448AC8DB5F8181B305281"
STUDENT_ONE_ID = "DC90F9C55C6142519C9CC9EFA1847922"
STUDENT_TWO_ID = "5A2AECC78C2E43FD9CC03CCE55D9806D"
PASSWORD = "NOTREAL"  # noqa: S105
USER_NAME = "TestUser"
SCHOOL_CODE = "88552"



"""Define a stop object."""

from datetime import datetime, time
from typing import Annotated, Self

from dateutil import parser
from lxml import etree
from pydantic import BaseModel, BeforeValidator

from . import parse_time_str, parse_yn_bool, xpath_element, xpath_elements

# Type aliases for cleaner field definitions
TimeStr = Annotated[time, BeforeValidator(parse_time_str)]
YNBool = Annotated[bool, BeforeValidator(parse_yn_bool)]


def _parse_float(value: str) -> float:
    """Parse float from string, defaulting to 0 for empty."""
    return 0.0 if value == "" else float(value)


def _parse_int(value: str) -> int:
    """Parse int from string, defaulting to 0 for empty."""
    return 0 if value == "" else int(value)


def _parse_datetime(value: str) -> datetime:
    """Parse datetime string."""
    return parser.parse(value)


FloatStr = Annotated[float, BeforeValidator(_parse_float)]
IntStr = Annotated[int, BeforeValidator(_parse_int)]
DateTimeStr = Annotated[datetime, BeforeValidator(_parse_datetime)]


class StudentStop(BaseModel):
    """Define a student stop."""

    name: str
    latitude: FloatStr
    longitude: FloatStr
    start_time: TimeStr
    stop_type: str
    substitute_vehicle_name: str
    vehicle_name: str
    stop_id: str
    arrival_time: TimeStr
    time_of_day_id: str
    vehicle_id: str
    esn: str
    tier_start_time: TimeStr
    bus_visibility_start_offset: IntStr

    @classmethod
    def from_element(cls, elem: etree._Element) -> "StudentStop":
        """Create from lxml element."""
        return cls(
            name=elem.get("Name", ""),
            latitude=elem.get("Latitude", "0"),
            longitude=elem.get("Longitude", "0"),
            start_time=elem.get("StartTime", "00:00:00"),
            stop_type=elem.get("StopType", ""),
            substitute_vehicle_name=elem.get("SubstituteVehicleName", ""),
            vehicle_name=elem.get("VehicleName", ""),
            stop_id=elem.get("StopId", ""),
            arrival_time=elem.get("ArrivalTime", "00:00:00"),
            time_of_day_id=elem.get("TimeOfDayId", ""),
            vehicle_id=elem.get("VehicleId", ""),
            esn=elem.get("Esn", ""),
            tier_start_time=elem.get("TierStartTime", "00:00:00"),
            bus_visibility_start_offset=elem.get("BusVisibilityStartOffset", "0"),
        )


class VehicleLocation(BaseModel):
    """Define a student vehicle location."""

    name: str
    latitude: FloatStr
    longitude: FloatStr
    log_time: DateTimeStr
    ignition: YNBool
    latent: YNBool
    time_zone_offset: int
    heading: str
    speed: int
    address: str
    message_code: int
    display_on_map: YNBool

    @classmethod
    def from_element(cls, elem: etree._Element) -> "VehicleLocation":
        """Create from lxml element."""
        return cls(
            name=elem.get("Name", ""),
            latitude=elem.get("Latitude", "0"),
            longitude=elem.get("Longitude", "0"),
            log_time=elem.get("LogTime", ""),
            ignition=elem.get("Ignition", "N"),
            latent=elem.get("Latent", "N"),
            time_zone_offset=int(elem.get("TimeZoneOffset", "0")),
            heading=elem.get("Heading", ""),
            speed=int(elem.get("Speed", "0")),
            address=elem.get("Address", ""),
            message_code=int(elem.get("MessageCode", "0")),
            display_on_map=elem.get("DisplayOnMap", "N"),
        )


class StopResponse(BaseModel):
    """Define a stop object."""

    vehicle_location: VehicleLocation | None
    student_stops: list[StudentStop]

    @classmethod
    def from_text(cls, response_text: str) -> Self:
        """Create a new instance from text."""
        root = etree.fromstring(response_text.encode())

        vehicle_elem = xpath_element(root, "//*[local-name()='VehicleLocation']")
        vehicle_location = (
            VehicleLocation.from_element(vehicle_elem)
            if vehicle_elem is not None
            else None
        )

        stop_elements = xpath_elements(root, "//*[local-name()='StudentStop']")
        student_stops = [StudentStop.from_element(e) for e in stop_elements]

        return cls(vehicle_location=vehicle_location, student_stops=student_stops)



"""Test the hcb soap client."""

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from hcb_soap_client.hcb_soap_client import HcbApiError, HcbSoapClient
from tests.test_data.const import (
    ACCOUNT_ID,
    PASSWORD,
    SCHOOL_CODE,
    SCHOOL_ID,
    STUDENT_ONE_ID,
    USER_NAME,
)

from . import read_file


@pytest.fixture
def mock_session() -> MagicMock:
    """Create a mock aiohttp session with common setup."""
    session = MagicMock()
    session.post.return_value.__aenter__.return_value.status = 200
    session.close = AsyncMock(return_value=None)
    return session


@pytest.fixture
def mock_session_with_response(mock_session: MagicMock) -> MagicMock:
    """Create a mock session factory that sets response text."""

    def _set_response(filename: str) -> MagicMock:
        mock_session.post.return_value.__aenter__.return_value.text.return_value = (
            read_file(filename)
        )
        return mock_session

    mock_session.set_response = _set_response
    return mock_session


@patch("hcb_soap_client.hcb_soap_client.aiohttp.ClientSession")
async def test_get_school_id(
    mock_client: MagicMock, mock_session_with_response: MagicMock
) -> None:
    """Tests the get school id."""
    session = mock_session_with_response.set_response("s1100.xml")
    mock_client.return_value.__aenter__.return_value = session
    client = HcbSoapClient()
    response = await client.get_school_id(SCHOOL_CODE)
    assert response == SCHOOL_ID


@patch("hcb_soap_client.hcb_soap_client.aiohttp.ClientSession")
async def test_get_parent_info(
    mock_client: MagicMock, mock_session_with_response: MagicMock
) -> None:
    """Tests the account response."""
    session = mock_session_with_response.set_response("s1157.xml")
    mock_client.return_value.__aenter__.return_value = session
    client = HcbSoapClient()
    response = await client.get_parent_info(SCHOOL_ID, USER_NAME, PASSWORD)
    assert response.account_id == ACCOUNT_ID
    expected_students = 2
    assert len(response.students) == expected_students


@patch("hcb_soap_client.hcb_soap_client.aiohttp.ClientSession")
async def test_get_stop_info(
    mock_client: MagicMock, mock_session_with_response: MagicMock
) -> None:
    """Tests the stop info response."""
    session = mock_session_with_response.set_response("s1158_AM.xml")
    mock_client.return_value.__aenter__.return_value = session
    client = HcbSoapClient()
    response = await client.get_stop_info(
        SCHOOL_ID, ACCOUNT_ID, STUDENT_ONE_ID, HcbSoapClient.AM_ID
    )
    assert response.vehicle_location is not None
    assert response.vehicle_location.address != ""
    expected_stops = 2
    assert len(response.student_stops) == expected_stops


def test_init() -> None:
    """Test the init."""
    client = HcbSoapClient()
    assert client._url == "https://api.synovia.com/SynoviaApi.svc"
    assert client._owns_session is True

    client = HcbSoapClient("http://test.url")
    assert client._url == "http://test.url"

    client = HcbSoapClient(None)
    assert client._url == "https://api.synovia.com/SynoviaApi.svc"

    # Test with provided session
    mock_session = MagicMock()
    client = HcbSoapClient(session=mock_session)
    assert client._session is mock_session
    assert client._owns_session is False


@patch("hcb_soap_client.hcb_soap_client.aiohttp.ClientSession")
async def test_context_manager(mock_client: MagicMock) -> None:
    """Test the async context manager creates and closes session."""
    mock_session = MagicMock()
    mock_session.close = AsyncMock(return_value=None)
    mock_client.return_value = mock_session

    async with HcbSoapClient() as client:
        assert client._session is mock_session
        mock_client.assert_called_once()

    mock_session.close.assert_called_once()
    assert client._session is None


async def test_context_manager_with_provided_session() -> None:
    """Test context manager does not close a provided session."""
    mock_session = MagicMock()
    mock_session.close = AsyncMock(return_value=None)

    async with HcbSoapClient(session=mock_session) as client:
        # Session should not be replaced
        assert client._session is mock_session

    # Should not close provided session
    mock_session.close.assert_not_called()


@patch("hcb_soap_client.hcb_soap_client.aiohttp.ClientSession")
async def test_get_school_id_with_session(
    mock_client: MagicMock, mock_session_with_response: MagicMock
) -> None:
    """Test API call using context manager with reused session."""
    session = mock_session_with_response.set_response("s1100.xml")
    mock_client.return_value = session

    async with HcbSoapClient() as client:
        response = await client.get_school_id(SCHOOL_CODE)
        assert response == SCHOOL_ID


@patch("hcb_soap_client.hcb_soap_client.aiohttp.ClientSession")
async def test_api_error_handling(mock_client: MagicMock) -> None:
    """Test that API errors are raised properly."""
    mock_session = MagicMock()
    mock_response = MagicMock()
    mock_response.status = 500
    mock_response.text = AsyncMock(return_value="Internal Server Error")
    mock_session.post.return_value.__aenter__.return_value = mock_response
    mock_client.return_value.__aenter__.return_value = mock_session

    client = HcbSoapClient()
    with pytest.raises(HcbApiError) as exc_info:
        await client.get_school_id(SCHOOL_CODE)

    assert exc_info.value.status_code == 500
    assert "500" in str(exc_info.value)


def test_hcb_api_error() -> None:
    """Test HcbApiError exception."""
    error = HcbApiError("Test error", status_code=404)
    assert str(error) == "Test error"
    assert error.status_code == 404

    error_no_status = HcbApiError("Another error")
    assert error_no_status.status_code is None



"""Test the stop response."""

from datetime import datetime, time

from hcb_soap_client.stop_response import StopResponse
from tests import read_file


def test_stop_response_from_dict_am() -> None:
    """Test the from text."""
    response_text = read_file("s1158_AM.xml")

    stop_response = StopResponse.from_text(response_text)
    assert stop_response.vehicle_location is not None
    assert stop_response.vehicle_location.name == "20-05"
    assert stop_response.vehicle_location.latitude == 34.7902536
    assert stop_response.vehicle_location.longitude == -86.782654999999991
    assert stop_response.vehicle_location.log_time == datetime(2024, 10, 31, 7, 0)
    assert not stop_response.vehicle_location.ignition
    assert not stop_response.vehicle_location.latent
    assert stop_response.vehicle_location.time_zone_offset == -1
    assert stop_response.vehicle_location.heading == "E "
    assert stop_response.vehicle_location.speed == 35
    assert stop_response.vehicle_location.address == "Endeavor Elementary School"
    assert stop_response.vehicle_location.message_code == 2
    assert not stop_response.vehicle_location.display_on_map
    assert len(stop_response.student_stops) == 2
    assert stop_response.student_stops[0].name == "Endeavor Elementary School (AM)"
    assert stop_response.student_stops[0].latitude == 34.790386
    assert stop_response.student_stops[0].longitude == -86.784177
    assert stop_response.student_stops[0].start_time == time(7, 0, 0)
    assert stop_response.student_stops[0].stop_type == "School"
    assert stop_response.student_stops[0].substitute_vehicle_name == ""
    assert stop_response.student_stops[0].vehicle_name == "20-05"
    assert (
        stop_response.student_stops[0].stop_id == "275FD0B6-3836-4FAE-A185-14C0C1ED861D"
    )
    assert stop_response.student_stops[0].arrival_time == time(6, 0, 0)
    assert (
        stop_response.student_stops[0].time_of_day_id
        == "55632A13-35C5-4169-B872-F5ABDC25DF6A"
    )
    assert (
        stop_response.student_stops[0].vehicle_id
        == "546DE805-C8DF-4464-B6FE-9DB2D05BA9DD"
    )
    assert stop_response.student_stops[0].esn == "123456"
    assert stop_response.student_stops[0].tier_start_time == time(5, 42, 0)
    assert stop_response.student_stops[0].bus_visibility_start_offset == 0
    assert stop_response.student_stops[1].name == "Circle K (AM)"
    assert stop_response.student_stops[1].latitude == 34.807730
    assert stop_response.student_stops[1].longitude == -86.749860
    assert stop_response.student_stops[1].start_time == time(7, 0, 0)
    assert stop_response.student_stops[1].stop_type == "Stop"
    assert stop_response.student_stops[1].substitute_vehicle_name == ""
    assert stop_response.student_stops[1].vehicle_name == "20-05"
    assert (
        stop_response.student_stops[1].stop_id == "CBD14468-96E3-4648-9DBA-F0B8687A02BC"
    )
    assert stop_response.student_stops[1].arrival_time == time(6, 5, 0)
    assert (
        stop_response.student_stops[1].time_of_day_id
        == "55632A13-35C5-4169-B872-F5ABDC25DF6A"
    )
    assert (
        stop_response.student_stops[1].vehicle_id
        == "546DE805-C8DF-4464-B6FE-9DB2D05BA9DD"
    )
    assert stop_response.student_stops[1].esn == "123456"
    assert stop_response.student_stops[1].tier_start_time == time(5, 42, 0)
    assert stop_response.student_stops[1].bus_visibility_start_offset == 0


def test_stop_response_from_dict_pm() -> None:
    """Test the from text."""
    response_text = read_file("s1158_PM.xml")
    stop_response = StopResponse.from_text(response_text)
    assert stop_response.vehicle_location is not None
    assert stop_response.vehicle_location.name == "20-05"
    assert stop_response.vehicle_location.latitude == 34.7902536
    assert stop_response.vehicle_location.longitude == -86.782654999999991
    assert stop_response.vehicle_location.log_time == datetime(2024, 10, 31, 7, 0)
    assert not stop_response.vehicle_location.ignition
    assert not stop_response.vehicle_location.latent
    assert stop_response.vehicle_location.time_zone_offset == -1
    assert stop_response.vehicle_location.heading == "E "
    assert stop_response.vehicle_location.speed == 35
    assert stop_response.vehicle_location.address == "Endeavor Elementary School"
    assert stop_response.vehicle_location.message_code == 2
    assert not stop_response.vehicle_location.display_on_map
    assert len(stop_response.student_stops) == 2
    assert stop_response.student_stops[0].name == "Endeavor Elementary School PM"
    assert stop_response.student_stops[0].latitude == 34.790386
    assert stop_response.student_stops[0].longitude == -86.784177
    assert stop_response.student_stops[0].start_time == time(7, 0, 0)
    assert stop_response.student_stops[0].stop_type == "School"
    assert stop_response.student_stops[0].substitute_vehicle_name == ""
    assert stop_response.student_stops[0].vehicle_name == "20-05"
    assert (
        stop_response.student_stops[0].stop_id == "9A361D75-F0BA-4F7D-94ED-BE6838BFAA6B"
    )
    assert stop_response.student_stops[0].arrival_time == time(6, 0, 0)
    assert (
        stop_response.student_stops[0].time_of_day_id
        == "6E7A050E-0295-4200-8EDC-3611BB5DE1C1"
    )
    assert (
        stop_response.student_stops[0].vehicle_id
        == "546DE805-C8DF-4464-B6FE-9DB2D05BA9DD"
    )
    assert stop_response.student_stops[0].esn == "123456"
    assert stop_response.student_stops[0].tier_start_time == time(5, 42, 0)
    assert stop_response.student_stops[0].bus_visibility_start_offset == 0
    assert stop_response.student_stops[1].name == "Circle K (PM)"
    assert stop_response.student_stops[1].latitude == 34.807730
    assert stop_response.student_stops[1].longitude == -86.749860
    assert stop_response.student_stops[1].start_time == time(7, 0, 0)
    assert stop_response.student_stops[1].stop_type == "Stop"
    assert stop_response.student_stops[1].substitute_vehicle_name == ""
    assert stop_response.student_stops[1].vehicle_name == "20-05"
    assert (
        stop_response.student_stops[1].stop_id == "D67ADB46-0D81-4E9F-9520-B5569098A9D2"
    )
    assert stop_response.student_stops[1].arrival_time == time(6, 5, 0)
    assert (
        stop_response.student_stops[1].time_of_day_id
        == "6E7A050E-0295-4200-8EDC-3611BB5DE1C1"
    )
    assert (
        stop_response.student_stops[1].vehicle_id
        == "546DE805-C8DF-4464-B6FE-9DB2D05BA9DD"
    )
    assert stop_response.student_stops[1].esn == "123456"
    assert stop_response.student_stops[1].tier_start_time == time(5, 42, 0)
    assert stop_response.student_stops[1].bus_visibility_start_offset == 0



"""Connect to HCB soap api."""

from types import TracebackType
from typing import Self

import aiohttp
from lxml import etree
from lxml.builder import ElementMaker

from . import xpath_attr
from .account_response import AccountResponse
from .stop_response import StopResponse

# App version used in headers and parameters
APP_VERSION = "3.6.0"

# Namespace definitions
SOAP_NS = "http://schemas.xmlsoap.org/soap/envelope/"
TEMPURI_NS = "http://tempuri.org/"

# Element makers for different namespaces
SOAP = ElementMaker(namespace=SOAP_NS, nsmap={"soap": SOAP_NS})
TEMPURI = ElementMaker(namespace=TEMPURI_NS, nsmap={None: TEMPURI_NS})

DEFAULT_HEADERS = {
    "app-version": APP_VERSION,
    "app-name": "hctb",
    "client-version": APP_VERSION,
    "user-agent": f"hctb/{APP_VERSION} App-Press/{APP_VERSION}",
    "cache-control": "no-cache",
    "content-type": "text/xml",
    "host": "api.synovia.com",
    "connection": "Keep-Alive",
    "accept-encoding": "gzip",
    "cookie": "SRV=prdweb1",
}


class HcbApiError(Exception):
    """Exception raised when the HCB API returns an error."""

    def __init__(self, message: str, status_code: int | None = None) -> None:
        """Initialize the exception."""
        super().__init__(message)
        self.status_code = status_code


def _build_soap_envelope(method: str, params: list[tuple[str, str]]) -> bytes:
    """Build a complete SOAP envelope with the given method and parameters."""
    # Build the method element with parameters
    param_elements = [TEMPURI(name, value) for name, value in params]
    method_element = TEMPURI(method, *param_elements)

    # Wrap in SOAP envelope
    envelope = SOAP.Envelope(SOAP.Body(method_element))

    return etree.tostring(envelope, xml_declaration=True, encoding="utf-8")


class HcbSoapClient:
    """Define soap client."""

    AM_ID = "55632A13-35C5-4169-B872-F5ABDC25DF6A"
    PM_ID = "6E7A050E-0295-4200-8EDC-3611BB5DE1C1"

    def __init__(
        self,
        url: str | None = None,
        session: aiohttp.ClientSession | None = None,
    ) -> None:
        """Create an instance of the client."""
        self._url = url or "https://api.synovia.com/SynoviaApi.svc"
        self._session = session
        self._owns_session = session is None

    async def __aenter__(self) -> Self:
        """Enter async context manager."""
        if self._owns_session:
            self._session = aiohttp.ClientSession()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit async context manager and close owned session."""
        if self._owns_session and self._session is not None:
            await self._session.close()
            self._session = None

    async def _request(self, method: str, params: list[tuple[str, str]]) -> str:
        """Make a SOAP request and return the response text."""
        payload = _build_soap_envelope(method, params)
        headers = {
            **DEFAULT_HEADERS,
            "soapaction": f"http://tempuri.org/ISynoviaApi/{method}",
        }

        # Use existing session or create a temporary one
        if self._session is not None:
            async with self._session.post(
                self._url, data=payload, headers=headers
            ) as response:
                await self._check_response(response)
                return await response.text()

        # Fallback: create session for single request (less efficient)
        async with (
            aiohttp.ClientSession() as session,
            session.post(self._url, data=payload, headers=headers) as response,
        ):
            await self._check_response(response)
            return await response.text()

    async def _check_response(self, response: aiohttp.ClientResponse) -> None:
        """Check the response for errors and raise if necessary."""
        if response.status >= 400:  # noqa: PLR2004
            text = await response.text()
            msg = f"API request failed with status {response.status}: {text[:200]}"
            raise HcbApiError(msg, status_code=response.status)

    async def get_school_id(self, school_code: str) -> str:
        """Return the school ID from the api."""
        response_text = await self._request("s1100", [("P1", school_code)])
        root = etree.fromstring(response_text.encode())
        return xpath_attr(root, "//*[local-name()='Customer']/@ID")

    async def get_parent_info(
        self, school_id: str, username: str, password: str
    ) -> AccountResponse:
        """Return the user info from the api."""
        params = [
            ("P1", school_id),
            ("P2", username),
            ("P3", password),
            ("P4", "LookupItem_Source_Android"),
            ("P5", "Android"),
            ("P6", APP_VERSION),
            ("P7", ""),
        ]
        response_text = await self._request("s1157", params)
        return AccountResponse.from_text(response_text)

    async def get_stop_info(
        self, school_id: str, parent_id: str, student_id: str, time_of_day_id: str
    ) -> StopResponse:
        """Return the bus stop info from the api."""
        params = [
            ("P1", school_id),
            ("P2", parent_id),
            ("P3", student_id),
            ("P4", time_of_day_id),
            ("P5", "true"),
            ("P6", "false"),
            ("P7", "10"),
            ("P8", "14"),
            ("P9", "english"),
        ]
        response_text = await self._request("s1158", params)
        return StopResponse.from_text(response_text)
